@@ -12,7 +12,7 @@
 	$NUMBER{"jperl"} = &JPERL;
 	$Japanese'NOCONV = $NUMBER{"jperl"};
 	$Japanese'OUT = $MSDOS ? "Shift_JIS" : "EUC-JP";
-	$RCSID = q$Id: htroff.pl,v 1.38 1998/10/09 03:33:05 toyoda Exp $;
+	$RCSID = q$Id: htroff.pl,v 1.39 1998/11/26 16:37:47 toyoda Exp toyoda $;
 	&initFindPath;
 
 	# common variables and messages
@@ -30,7 +30,8 @@
 	@MACROARG = ();		# macro arguments: for \\$n
 	$DEFMACRO = undef;	# if TRUE, lines are stored as macro
 	%MACROS = ();		# LF separated macro lines
-	%ALIAS = ();		# macro aliases
+	%ALIAS = ('cleanup', 'nop', 'opened', 'nop');
+		# request/macro aliases
 
 	# pattern-matching request hook
 	%HOOK = ('^\s*$', 'tag P');
@@ -46,6 +47,9 @@
 	@INPUT = ();
 	$STRING{"<"} = $INPUT = "";
 	$STRING{">"} = $OUTPUT = "STDOUT";
+	# for better diag
+	%LINE_COUNTER = ();
+	$MACROS_DONE = 0;
 
 	# predefined number registers
 	($NUMBER{"sec"}, $NUMBER{"min"}, $NUMBER{"hour"}, $NUMBER{"dy"},
@@ -71,6 +75,7 @@
 		/^-(.)/;
 		$OPTION{$1} = $';
 	}
+	push(@INPUT, 'end of macros');
 	push(@INPUT, @ARGV);
 
 	# event loop
@@ -84,6 +89,7 @@
 		&ProcessLine;
 		$CONTINUE = "";
 	}
+	&Request(".cleanup");
 exit 0;
 
 sub RegisterMacro {
@@ -372,14 +378,28 @@ sub Request {
 	} elsif ($reqname eq "rm") {
 		$MACROS{$REQARG[0]} = undef;
 	} elsif ($reqname eq "nr") {
+		$numreg_name = shift(@REQARG);
+		$NUMBER{$numreg_name} = shift(@REQARG);
+		$NUMBER{"+$numreg_name"} = shift(@REQARG);
+	} elsif ($reqname eq "nrexpr") {
 		$numreg_name = shift(@REQARG);
 		$NUMBER{$numreg_name} = &ExprList(@REQARG);
 	} elsif ($reqname eq "ds") {
+		$strreg_name = shift(@REQARG);
+		$STRING{$strreg_name} = join(" ", @REQARG);
+		$STRING{$strreg_name} =~ s/\0//g;
+	} elsif ($reqname eq "dsexpr") {
 		$strreg_name = shift(@REQARG);
 		$STRING{$strreg_name} = &ExprList(@REQARG);
+		$STRING{$strreg_name} =~ s/\0//g;
 	} elsif ($reqname eq "as") {
+		$strreg_name = shift(@REQARG);
+		$STRING{$strreg_name} .= join(" ", @REQARG);
+		$STRING{$strreg_name} =~ s/\0//g;
+	} elsif ($reqname eq "asexpr") {
 		$strreg_name = shift(@REQARG);
 		$STRING{$strreg_name} .= &ExprList(@REQARG);
+		$STRING{$strreg_name} =~ s/\0//g;
 	} elsif ($reqname eq "bp") {
 		&tag("HR");
 	} elsif ($reqname eq "br") {
@@ -402,14 +422,13 @@ sub Request {
 		open($INPUT, "<$INPUT") || warn "$ME <$INPUT $CO\n";
 	} elsif ($reqname eq "ce") {
 		&SetCenter($REQARG[0] || 1);
-	# --- I don't remember whether they are in troff.
-	} elsif ($reqname eq "it") {
-		&write(&SetFont("I", ((@REQARG > 0) ? $REQARG[0] : 1)));
 	} elsif ($reqname eq "ul") {
 		&write(&SetFont("U", ((@REQARG > 0) ? $REQARG[0] : 1)));
+	# --- htroff original requests
+	} elsif ($reqname eq "it") {
+		&write(&SetFont("I", ((@REQARG > 0) ? $REQARG[0] : 1)));
 	} elsif ($reqname eq "bf") {
 		&write(&SetFont("B", ((@REQARG > 0) ? $REQARG[0] : 1)));
-	# --- htroff original requests
 	} elsif ($reqname eq "shift") {
 		shift(@MACROARG);
 	} elsif ($reqname eq "getline") {
@@ -422,8 +441,12 @@ sub Request {
 		&SetOutput($REQARG[0]);
 	} elsif ($reqname eq "append") {
 		&SetAppend($REQARG[0]);
+	} elsif ($reqname eq "keep") {
+		$STRING{"keep"} = $REQARG[0];
+		delete $STRING{"keep"} if (@REQARG < 1);
 	} elsif ($reqname eq "alias") {
 		$ALIAS{$REQARG[0]} = $REQARG[1];
+		delete $ALIAS{$REQARG[0]} if (@REQARG < 2);
 	} elsif ($reqname eq "hook") {
 		$REQARG[0] =~ s/\0//g;
 		$HOOK{$REQARG[0]} = $REQARG[1];
@@ -510,13 +533,17 @@ sub SetAppend {
 }
 
 sub NextFile {
+	close($INPUT) unless ($INPUT eq "DATA");
 	for (;;) {
 		if (@PUSHED_INPUT) {
 			return ($STRING{"<"} = $INPUT = pop(@PUSHED_INPUT));
 		}
 		return undef unless (@INPUT);
 		$STRING{"<"} = $INPUT = shift(@INPUT);
-		if ($INPUT =~ /^DATA::(.*)/) {
+		if ($INPUT eq 'end of macros') {
+			$MACROS_DONE = 1;
+			next;
+		} elsif ($INPUT =~ /^DATA::(.*)/) {
 			$macroname = $1;
 			$STRING{"<"} = $INPUT = "DATA";
 			# skip operation
@@ -526,7 +553,11 @@ sub NextFile {
 			}
 			warn "$ME macro package $macroname $NF.\n";
 		} else {
-			open($INPUT, "<$INPUT") && return $INPUT;
+			open($INPUT, "<$INPUT") && do {
+				&Request(".opened $INPUT") if $MACROS_DONE;
+				$LINE_COUNTER{$INPUT} = 1;
+				return $INPUT;
+			};
 			warn "$ME <$INPUT $CO\n";
 		}
 	}
@@ -535,6 +566,7 @@ sub NextFile {
 sub getline {
 	for (;;) {
 		$getline = <$INPUT>;
+		$NUMBER{"lines"} = $LINE_COUNTER{$INPUT}++;
 		next if (($INPUT eq "DATA") && ($getline =~ /^DATA/));
 		$getline = ""
 			if (($INPUT eq "DATA") && ($getline =~ /^ENDDATA/));
@@ -568,15 +600,22 @@ sub ExpandEscape {
 	s/\\\$\+/&MacroArgAll/ge;
 	s/\\\$([1-9])/$MACROARG[$1-1]/g;
 
-	# number register substitution 
-	s/\\n\[([^]]+)]/$NUMBER{$1}+0/ge;
-	s/\\n\((..)/$NUMBER{$1}+0/ge;
-	s/\\n(.)/$NUMBER{$1}+0/ge;
-
 	# string register substitution
 	s/\\\*\[([^]]+)]/$STRING{$1}/g;
 	s/\\\*\((..)/$STRING{$1}/g;
 	s/\\\*(.)/$STRING{$1}/g;
+
+	# inline size change --- ignored
+	s/\\s[-+]?[0-9]+/\0/g;
+
+	# number register substitution 
+	s/\\n\+\[([^]]+)]/$NUMBER{$1} += $NUMBER{"+$1"}/ge;
+	s/\\n\[([^]]+)]/$NUMBER{$1}+0/ge;
+	s/\\n\+\((..)/$NUMBER{$1} += $NUMBER{"+$1"}/ge;
+	s/\\n\((..)/$NUMBER{$1}+0/ge;
+	s/\\n\+(.)/$NUMBER{$1} += $NUMBER{"+$1"}/ge;
+	s/\\n(.)/$NUMBER{$1}+0/ge;
+
 }
 
 sub write {
@@ -598,18 +637,38 @@ sub write {
 			s/\\fR/&SetFont("", 0)/e;
 		}
 	}
-	
-	# inline size change --- ignored
-	s/\\s-?[0-9]+//g;
+
+	# upward or downward shift
+	$UPDOWN = 0;
+	while (/\\[ud]/) {
+		if ($& eq "\\u") {
+			$UPDOWN++;
+			if ($UPDOWN > 0) { s/\\u/\\QLSUP\\QG/; }
+			else { s/\\u/\\QL\/SUB\\QG/; }
+		} elsif ($& eq "\\d") {
+			$UPDOWN--;
+			if ($UPDOWN < 0) { s/\\d/\\QLSUB\\QG/; }
+			else { s/\\d/\\QL\/SUP\\QG/; }
+		}
+	}
+	while ($UPDOWN > 0) {
+		s/$/\\QL\/SUP\\QG/;
+		$UPDOWN--;
+	}
+	while ($UPDOWN < 0) {
+		s/$/\\QL\/SUB\\QG/;
+		$UPDOWN++;
+	}
 
 	# special one-char sequences
 	s/\\\&/\0/g;
 	s/\\\|/\0/g;
-	s/\\\^/ /g;
+	s/\\\^//g;		# quite narrow space ... better than one SP?
 	s/\\l/ /g;
 	s/\\-/-/g;
 	s/\\ /\\[nbsp]/g;
 	s/\\0/\\[nbsp]/g;
+	s/\\c$// && ($UNIFY_ONE_LINE = 1);	# continuation mark
 
 	# defined character escape
 	s/\\\((..)/$CHAR{$1}/g;
@@ -633,8 +692,13 @@ sub write {
 
 sub writeln {
 	local($line) = @_;
-	&write($line);
-	print "\n" unless $NUMBER{"unify"};
+	if (defined $STRING{'keep'}) {
+		$MACROS{$STRING{'keep'}} .= "$line\n";
+	} else {
+		&write($line);
+		print "\n" unless ($NUMBER{"unify"} || $UNIFY_ONE_LINE);
+		$UNIFY_ONE_LINE = undef;
+	}
 }
 
 sub CheckTag {
